"""
Definition of metadata model for input to and output of simtools.
Follows CTA top-level data model definition.
Metadata is broken into hierarchical categories in this module.

Metadata definitions require the three following fields to be defined:

* type: data type
* required: boolean if this field is required as input(!)
* default: default value

These definitions are used for:

* data products submitted to SimPipe ('input')
* data products generated by SimPipe ('output')

"""


class InvalidSchemaList(Exception):
    """
    Exception raised for requests of unknown schema lists
    """


def _category_reference():
    """
    Metadata field REFERENCE.

    Refers to the CTA Top-Level Data Model (document CTA-SPE-OSO-000000-0001.)

    """
    return {"VERSION": {"type": "str", "required": True, "default": "1.0.0"}}


def _category_contact():
    """
    Metadata field CONTACT.

    Organisation and person submitting or processing these data.

    """
    return {
        "ORGANIZATION": {"type": "str", "required": True, "default": None},
        "NAME": {"type": "str", "required": True, "default": None},
        "EMAIL": {"type": "email", "required": True, "default": None},
    }


def _category_product():
    """
    Metadata field PRODUCT.

    Describes the data product, especially its validity and associations.

    """
    return {
        "DESCRIPTION": {"type": "str", "required": True, "default": None},
        "CREATION_TIME": {"type": "datetime", "required": True, "default": None},
        "ID": {"type": "str", "required": False, "default": None},
        "DATA": {
            "CATEGORY": {"type": "str", "required": False, "default": None},
            "LEVEL": {"type": "str", "required": False, "default": None},
            "TYPE": {"type": "str", "required": False, "default": None},
            "MODEL": {
                "NAME": {"type": "str", "required": False, "default": None},
                "VERSION": {"type": "str", "required": False, "default": None},
                "URL": {"type": "str", "required": False, "default": None},
                "TYPE": {"type": "str", "required": False, "default": None},
                "SUBTYPE": {"type": "str", "required": False, "default": None},
            },
        },
        "FORMAT": {"type": "str", "required": False, "default": "ecsv"},
        "VALID": {
            "START": {"type": "datetime", "required": False, "default": None},
            "END": {"type": "datetime", "required": False, "default": None},
        },
        "ASSOCIATION": {"type": "instrumentlist", "required": True, "default": []},
    }


def _category_process():
    """
    Metadata field PROCESS.

    Description of the process which generated these data.

    """
    return {
        "TYPE": {"type": "str", "required": True, "default": None},
        "SUBTYPE": {"type": "str", "required": False, "default": None},
        "ID": {"type": "str", "required": True, "default": None},
    }


def _category_context():
    """
    Metadata field CONTEXT.

    Describes list of context documents.

    """
    return {"DOCUMENT": {"type": "documentlist", "required": False, "default": None}}


def _category_document():
    """
    Metadata field DOCUMENT.

    """

    return {
        "TYPE": {"type": "str", "required": False, "default": None},
        "ID": {"type": "str", "required": False, "default": None},
        "LINK": {"type": "str", "required": False, "default": None},
        "TITLE": {"type": "str", "required": False, "default": None},
    }


def _category_instrument():
    """
    Metadata field INSTRUMENT.

    Describes the instrument used to obtain these data or the instrument these data is applied for.

    """
    return {
        "SITE": {"type": "str", "required": True, "default": None},
        "CLASS": {"type": "str", "required": True, "default": None},
        "TYPE": {"type": "str", "required": True, "default": None},
        "SUBTYPE": {"type": "str", "required": False, "default": None},
        "ID": {"type": "str", "required": True, "default": None},
    }


def _category_activity():
    """
    Metadata field ACTIVITY.

    Describes the software used to process these data.

    """
    return {
        "NAME": {"type": "str", "required": True, "default": None},
        "TYPE": {"type": "str", "required": True, "default": "software"},
        "ID": {"type": "str", "required": True, "default": None},
        "START": {"type": "datetime", "required": False, "default": None},
        "SOFTWARE": {
            "NAME": {"type": "str", "required": True, "default": None},
            "VERSION": {"type": "str", "required": True, "default": None},
        },
    }


def top_level_reference_schema():
    """
    Reference schema following the CTA Top-Level Data Model.

    This metadata schema is used for simtools data products.

    Returns
    -------
    dict with reference schema


    """

    _ref_schema = {
        "CTA": {
            "REFERENCE": _category_reference(),
            "CONTACT": _category_contact(),
            "PRODUCT": _category_product(),
            "INSTRUMENT": _category_instrument(),
            "PROCESS": _category_process(),
            "ACTIVITY": _category_activity(),
            "CONTEXT": {
                "SIM": {
                    "ASSOCIATION": [
                        _category_instrument(),
                    ],
                    "DOCUMENT": [
                        _category_document(),
                    ],
                },
            },
        }
    }
    _ref_schema = _metadata_dict_with_defaults(_ref_schema)
    return _remove_empty_lists(_ref_schema)


def metadata_input_reference_schema():
    """
    Reference data model scheme for input metadata.
    Describes metadata provided for input to simtools applications.

    Returns
    -------
    dict with input reference schema

    """

    return {
        "REFERENCE": _category_reference(),
        "CONTACT": _category_contact(),
        "PRODUCT": _category_product(),
        "INSTRUMENT": _category_instrument(),
        "PROCESS": _category_process(),
        "CONTEXT": _category_context(),
    }


def metadata_input_reference_document_list(schema_list):
    """
    Reference model data for input metata data of type documentlist or instrumentlist

    Parameters
    ----------
    schema_list: str
        List type to be returned (e.g., instrumentlist or documentlist)

    Returns
    -------
    dict with input reference schema for documentlist or instrumentlist

    """
    if schema_list.lower() == "instrumentlist":
        return _category_instrument()
    if schema_list.lower() == "documentlist":
        return _category_document()

    msg = f"Invalid schema list: {schema_list}"
    raise InvalidSchemaList(msg)


def _metadata_dict_with_defaults(meta_dict):
    """
    Prepare dictionary with default values filled and removal of all type/required/default dicts.

    Parameters
    ----------
    meta_dict
        Metadata dictionary

    Returns
    -------
    dict
        Metadata dictionary with default values filled

    """

    for key, value in meta_dict.items():
        if isinstance(value, dict):
            if value.keys() >= {"type", "required", "default"}:
                meta_dict[key] = value["default"]
            else:
                _metadata_dict_with_defaults(value)
        elif isinstance(value, list):
            for list_entry in value:
                _metadata_dict_with_defaults(list_entry)
        else:
            msg = f"Invalid schema list with missing type, required, or default fields: {key}"
            raise InvalidSchemaList(msg)
    return meta_dict


def _remove_empty_lists(meta_dict):
    """
    Remove entries of type list with length zero, as those are not used for output.

    """
    _entries_to_pop = []
    for key, value in meta_dict.items():
        if isinstance(value, dict):
            _remove_empty_lists(value)
        elif isinstance(value, list) and len(value) == 0:
            _entries_to_pop.append(key)
    for key in _entries_to_pop:
        meta_dict.pop(key)

    return meta_dict
